//
// Turtle Graphics Library extension for teaching OO concepts
// by Hideki Tsuiki, Kyoto University
// as a modification of Tatsuya Hagino's Turtle Graphics Library
// Copyright (C) 2000, Hideki Tsuiki, Kyoto University

// Turtle Graphics Library for Information Processing I
// Copyright (C) 1998, Tatsuya Hagino, Keio University
//
// Permission to use, copy, modify, and distribute this software
// for educational purpose only is hereby granted, provided that
// the above copyright notice appear in all copies and that both
// the copyright notice and this permission notice appear in
// supporting documentation.  This software is provided "as is"
// with no warranty.


import java.awt.*;
import java.awt.event.*;
import java.io.*;
import java.lang.Math;
import java.util.Vector;

/** ȥΥ饹 */
public class Turtle{
  /** ⤷Υեɤͤ true ʤ顢
   * ѤԤfalse ʤ顢ɽ®Ԥ
   * ͤ true*/
  boolean withKame = true;

  /** ⤷false ʤ顢ġΥȥ
   * withKame ͤ˴ؤ餺ƤΥȥ˹®Ԥ
   * ͤ true*/
  public static boolean withKameAll = true;

  public Color kameColor = Color.green;

  // animation parameters
  static double rotateStep = 5.0 *Math.PI/180.0 ;
  static int moveStep = 5;
  static int[] speedStep = {100000, 20, 5, 2};

  // turtle animation speed constants
  final static int speedNoKame = 0;
  final static int speedFast = 1;
  final static int speedNormal = 2;
  final static int speedSlow = 3;

  // set the menu
  static void speedAll(int s) {
    if (s < 0 || s >= speedStep.length) return;
    rotateStep = speedStep[s] *Math.PI/180.0;
    moveStep = speedStep[s];
    withKameAll = (s > 0);
  }


  TurtlePanel f;  // set by TurtlePanel
  double angle;		// turtle current angle 
  double x, y;		// turtle current position
  double dx, dy;	// dx = sin(angle),  dy = -cos(angle)
  boolean penDown;	// pen status (up or down)
  Color c = Color.black; // pen color
  int kameType = 0;
  // turtle animation rubber line
  int rx, ry;
  boolean rubber = false;
  int moveWait = 20;
  int rotateWait = 20;

  /**  (x, y) Ȥɸ angle γ٤ǡTurtle  */
  public Turtle(int x,int y, int ia)
  {
    this.x = ((double)x + 0.5);
    this.y = ((double)y + 0.5);
    setangle((double)ia *Math.PI/180.0);
    penDown = true;
  }

  /** (200,200) Ȥɸ 0 ٤γ٤ǡTurtle  */
  public Turtle()
  {
    this(200,200, 0);
  }

  // turtle
  static final int kameFig[][] =
    { {-12,-6,-12,6,0,18,12,6,12,-6,0,-18,-12,-6},
      {-18,-12,-12,-6},
      {-6,-24,0,-18,6,-24},
      {12,-6,18,-12},
      {12,6,18,12},
      {-6,24,0,18,6,24},
      {-18,12,-12,6},
      {-18,12,-18,-12,-6,-24,6,-24,18,-12,18,12,6,24,-6,24,-18,12},
      {-15,-15,-18,-24,-9,-21},
      {9,-21,18,-24,15,-15},
      {15,15,18,24,9,21},
      {-9,21,-18,24,-15,15},
      {-3,24,0,30,3,24},
      {-6,-24,-12,-36,-6,-48,6,-48,12,-36,6,-24}
    };
  static final int kameRFig[][] =
    { {-12,-6,-12,6,0,18,12,6,12,-6,0,-18,-12,-6},
      {-18,-12,-12,-6},
      {-6,-24,0,-18,6,-24},
      {12,-6,18,-12},
      {12,6,18,12},
      {-6,24,0,18,6,24},
      {-18,12,-12,6},
      {-18,12,-18,-12,-6,-24,6,-24,18,-12,18,12,6,24,-6,24,-18,12},
      {-15,-15,-24,-18,-9,-21},
      {-9,21,-24,18,-15,15},
      {-3,24,-3,30,3,24},
      {-6,-24,-6,-36,0,-48,12,-48,18,-36,6,-24},
      {9,-21,18,-30,15,-15},
      {15,15,18,30,9,21}
    };
  static final int kameLFig[][] =
    { {-12,-6,-12,6,0,18,12,6,12,-6,0,-18,-12,-6},
      {-18,-12,-12,-6},
      {-6,-24,0,-18,6,-24},
      {12,-6,18,-12},
      {12,6,18,12},
      {-6,24,0,18,6,24},
      {-18,12,-12,6},
      {-18,12,-18,-12,-6,-24,6,-24,18,-12,18,12,6,24,-6,24,-18,12},
      {-15,-15,-18,-30,-9,-21},
      {-9,21,-18,30,-15,15},
      {-3,24,3,30,3,24},
      {-6,-24,-18,-36,-12,-48,0,-48,6,-36,6,-24},
      {9,-21,24,-18,15,-15},
      {15,15,24,18,9,21}
    };
  public double kameScale = 0.4;
  public int kame[][] = kameFig;
  public int kameR[][] = kameRFig;
  public int kameL[][] = kameLFig;

  // draw animation turtle
  void kameDraw(Graphics g, int data[][]) {
    int ix = (int)x, iy = (int)y;
    g.setColor(kameColor);
    for (int i = 0; i < data.length; i++) {
      int px = 0, py = 0;
      for (int j = 0; j < data[i].length; j += 2) {
	int kx = data[i][j], ky = data[i][j+1];
	int nx = (int)((kx*(-dy) + ky*(-dx)) * kameScale);
	int ny = (int)((kx*dx+ ky*(-dy)) * kameScale);
        if (j > 0) g.drawLine(ix + px, iy + py, ix + nx, iy + ny);
	px = nx;
	py = ny;
      }
    }
    g.setColor(c);
    g.fillOval(ix - 1, iy - 1, 2,2);
  }

  void setFrame(TurtlePanel f)
  {
    this.f = f;
  }

  void show(Graphics g)
  {
    
    if (rubber) {
      g.setColor(c);
      g.drawLine(rx, ry, (int)x, (int)y);
    }
    if (withKame && withKameAll) {
      switch ((kameType/2) % 4) {
      case 0: case 2: 
	kameDraw(g, kame);
	break;
      case 1:
	kameDraw(g, kameR);
	break;
      case 3:
	kameDraw(g, kameL);
	break;
      }
    }
  }

  void fcheck()
  {
    if(f == null) {
      System.out.println("You need to add a turtle to a frame befor you use it!!");
      System.exit(1);
    }
  }

  // turtle animation update
  void kameShow(int wait) {
    kameType++;
    f.repaint();
    if (withKame && withKameAll) {
      try {
	Thread.sleep(wait);
      } catch (InterruptedException e) {
      }
    }
  }

  /** n ˿ʤࡣ */
  public void fd(int n) {
    double xx = x;
    double yy = y;
    int back = 1;
    if(n < 0){
      back = -1; n = -n;
    }
    fcheck();
    if (penDown) {
      rx = (int)x;
      ry = (int)y;
      rubber = true;
    }
    for (int i = moveStep; i < n; i += moveStep) {
      x = xx + back * dx * i;
      y = yy + back * dy * i;
      kameShow(moveWait);
    }
    x = xx + back * dx * n;
    y = yy + back * dy * n;
    if (penDown) {
      f.addLineElement((int)xx, (int)yy, (int)x, (int)y, c);
      rubber = false;
    }
    kameShow(moveWait);
  }

  /** n ˿ʤࡣ */
  public void bk(int n) {
    fd(-n);
  }

  void setangle(double a)
  {
    angle = a;
    dx = Math.sin(a);
    dy = -Math.cos(a);
  }

  // set turtle angle
  void kameAngle(double a) {
    dx = Math.sin(a);
    dy = -Math.cos(a);
    kameShow(rotateWait);
  }

  /** n ٤˲롣 */
  public void rt(int ia) {
    rtd((double)ia*Math.PI/180.0);
  }
  
  void rtd(double a)  {
    fcheck();
    for (double i = rotateStep; i < a; i += rotateStep)
      kameAngle(angle + i);
    angle = (angle + a);
    kameAngle(angle);
  }

  /** n ٤˲롣 */
  public void lt(int ia) {
    ltd((double)ia*Math.PI/180.0);
  }
  
  void ltd(double a)  {
    fcheck();
    for (double i = rotateStep; i < a; i += rotateStep)
      kameAngle(angle - i);
    angle = (angle - a);
    kameAngle(angle);
  }

  /** ڥ򤢤롣  */
  public void up() {
    penDown = false;
  }

  /** ڥ򤪤ڥ򤪤֤ǿʤȡεפ̤˻Ĥ롣*/
  public void down() {
    penDown = true;
  }

  /** ڥ󤬾夬äƤ뤫ɤ */
  public boolean isDown() {
    return (penDown);
  }

  /**  ڥο nc ѹ롣 */
  public void setColor(Color nc) {
    c = nc;
  }

  /** ưᤵ x ꤹ롣x = 20 ǥեȤǤ롣ۤᤤ*/
  public void speed(int x)
  {
    moveWait = x;
    rotateWait = x;
  }
  
  /** ߤκɸ X ʬ֤ */
  public int getX()
  {
    return (int)this.x;
  }

  /** ߤκɸ Y ʬ֤*/
  public int getY()
  {
    return (int)this.y;
  }

  /** ߤγ٤֤*/
  public int getAngle()
  {
    return (int)(this.angle * 180.0 /Math.PI);
  }

  /*  double ǽʬAPI λͤǤ double 
   * ԼʤΤǡint  moveTo 
   */
public int moveTo(double x, double y, double aa) { 
    double prevx = this.x; double prevy = this.y;
    //    double a = Math.atan2(-( y - prevy), x - prevx);
    double a = Math.atan2(x - prevx, -( y - prevy));
    setangle(a);
    int r = (int)Math.sqrt((prevx - x) * (prevx - x) + (prevy - y) * (prevy - y));
    fd(r);
    this.x = ((double)x + 0.5);
    this.y = ((double)y + 0.5);
    setangle(aa);
    kameShow(moveWait);    
    return(r);
  }

  /** (x, y) Ȥɸޤǰư롣ڥ򤪤֤ʤ顢褵롣
   * ưΥ֤
   */
  public int moveTo(int x, int y, int ia)
  {
    return moveTo((double)x, (double)y, (double)ia*Math.PI/180.0);
  }

  /* t Ʊɸޤǰư롣ڥ򤪤֤ʤ顢褵롣
   *   ưΥ֤
   */
  public int moveTo(Turtle t)
  {
    return moveTo(t.x, t.y, t.angle);
  }

}









